/* Use image data to find and apply guiding feedback     */
/* Functions:                                            */
/*   Find a single guide star within the image subarea   */
/*   Center star in the subarea                          */
/*   Write residual error to a log file                  */

int GuideImage(unsigned short *data,
  int width, int height,
  int guide, int guide_flag, int guide_log_flag)
{ 
  int i,j,k;
  unsigned short pixdata;
  double dataval;
  double bkgval;
  double anbkg;
  int nbkg;
  int guide_n;
  double x_err, y_err;
  double x_0, y_0;
  double *guide_star;
  FILE *fp_log; 
  
  guide_n = width*height;
  guide_star = (double *) malloc((guide_n)*sizeof(double));
  
  /* Average perimeter for a background estimate */
  /* Corners are weighted by 2 */
  
  k = 0;
  nbkg = 0;
  bkgval = 0.;
  for (j = 0; j < height; j++)
  {
     for (i = 0; i < width; i++)
    {
      pixdata = *(data + k);
      dataval = (double) pixdata;
      if ( (j == 0) || (j == (height - 1)) ) 
      {
        bkgval = bkgval + dataval;
        nbkg++;
      }
      if ( (i == 0) || (i == (width - 1))) 
      {
        bkgval = bkgval + dataval;
        nbkg++;
      }       
      guide_star[k] = dataval;
      k++;
    }
  } 

  anbkg = (double) nbkg;
  bkgval = bkgval / anbkg;
  
  /* Subtract the background */
  
  k = 0;
  for (j = 0; j < height; j++)
  {
    for (i = 0; i < width; i++)
    {
      guide_star[k] = guide_star[k] - bkgval;
      k++;
    }
  }
      
  /* Find the guide star in this new image */
  
  find_guide_star(guide_star, width, height, &x_err, &y_err);
  
  /* Find the error from the center of the image subarea  */
    
  x_0 = (double) width;
  x_0 = x_0/2.;
  y_0 = (double) height;
  y_0 = y_0/2.;
  
  x_err = x_err - x_0;
  y_err = y_err - y_0; 
  
  printf("jfk: %lf %lf\n", x_err, y_err);
    
  /* Use the latest error to correct the image */
  
  if ( guide_flag == TRUE )
  {
    if ( guide == TRACKING )
    {
      /* Correct telescope pointing based on guide star location */

      feedback_tracking(x_err, y_err);
    }
    else if ( guide == SELFTRACKING )
    {
      /* Correct telescope pointing based on guide star location */

      feedback_tracking(x_err, y_err);
    }
    else if  ( guide == OPTICAL )
    {
      /* Correct adaptive optics based on guide star location */
      
      feedback_ao(x_err, y_err);
    }
    else if  ( guide == SELFOPTICAL )
    {
      /* Correct adaptive optics based on guide star location */
      
      feedback_ao(x_err, y_err);
    }
  }
  
  if ( guide_log_flag == TRUE )
  {
    time_t timebuf;
    char *gmtstr;
    
    time(&timebuf);
    gmtstr = asctime(gmtime(&timebuf));
    if (gmtstr[strlen(gmtstr) - 1] == '\n') gmtstr[strlen(gmtstr) - 1]= '\0';
    fp_log=fopen("guide.log", "a");
    fprintf(fp_log,"%s  %f  %f\n", gmtstr, x_err, y_err);
    fflush(fp_log);
    fclose(fp_log);
  }
  
  free(guide_star);
  return(1);
}
   
/* Determine barycenter of guide star in image subarray */

void find_guide_star(double *data, int w, int h, double *x_s, double *y_s)
{
  int i,j,k;
  double xval,yval,pixval,sumval,xav,yav;
    
  k = 0;
  sumval = 0.;
  xav = 0.;
  yav = 0.;
  for (j = 0; j < h; j++)
  {
    yval = (double) j;
    for (i = 0; i < w; i++)
    {
      pixval = data[k];
      
      if ( pixval >= guide_threshold )
      {
        xval = (double) i;
        sumval = sumval + pixval;
        xav = xav + xval * pixval;
        yav = yav + yval * pixval;
      }
      k++;
    }
  }
    
  /* Handle frame without signal */ 
  
  if (sumval <= 0.)
  {
    *x_s = ((double) w)/2.; 
    *y_s = ((double) h)/2.;
    return;
  }
     
  xav = xav / sumval;
  yav = yav / sumval;
  *x_s = xav;
  *y_s = yav;
    
  return;
}



/* Acquire a self-guided image frame and save it as a fits file */
/* Uses auxiliary CCD to find and apply guiding corrections     */

int CaptureGuide(int *phase,
  unsigned short *image_data, 
  unsigned short *track_data,
  double exposure_i, int subarea_i, 
  int x_i, int y_i, int width_i, int height_i,
  double exposure_t, int subarea_t, 
  int x_t, int y_t, int width_t, int height_t,
  int guide, int guide_flag, int guide_log_flag)
{
  int err;
  int i;
  
  static int guide_phase;
  static int guide_x;
  static int guide_y;
  static int guide_w;
  static int guide_h; 
  static int guide_n;
  
       
  ccd_phase = *phase;   
  
  if ( ccd_phase == 2 )
  {
    /* End both track and image exposures without readout */

    eep.ccd = CCD_IMAGING;
    mcp.fanEnable = TRUE;           /* Fan on */
    mcp.shutterCommand = 2;         /* Shutter closed */
    mcp.ledState = 0;               /* LED off */
    err=SBIGUnivDrvCommand(CC_MISCELLANEOUS_CONTROL, &mcp, NULL);

    eep.ccd = CCD_IMAGING;
    err=SBIGUnivDrvCommand(CC_END_EXPOSURE, &eep, NULL);
    eep.ccd = CCD_TRACKING;
    err=SBIGUnivDrvCommand(CC_END_EXPOSURE, &eep, NULL);
    ccd_phase = 0;
    *phase = ccd_phase;
    if (err != CE_NO_ERROR)  
    {
      fprintf(stderr,"Error ending guided exposure");
      return(0);
    }
    return(1);
  } 

  /* Test that this is indeed a guided image request */
  /* If not, then use another routine */

  if ( guide == UNGUIDED )
  {
    fprintf(stderr,"Request for unguided exposure made to CaptureGuide");
    fprintf (stderr,"Use CaptureImage instead");
    return(0);
  } 
  
  /* Test that the camera has an internal or external track ccd */
  
  if ( (ccd_tracking == FALSE) && (ccd_ext_tracking == FALSE) )
  {
    fprintf(stderr,"Request for self-guided exposure with unsupported hardware ");
    fprintf (stderr,"Use CaptureImage instead");
    return(0);
  }    

  /* Test that guide feedback type is appropriate */

  if ( ( guide != SELFTRACKING ) && ( guide != SELFOPTICAL ) )
  {
    fprintf(stderr,"Request for unknown guide type %d made to CaptureGuide\n", guide);
    return(0);
  }

  if ( ccd_phase == 0 )
  { 
    /* Start image exposure */
  
    sep.ccd = CCD_IMAGING;
    sep.abgState = ABG_LOW7;
    sep.openShutter =  SC_OPEN_SHUTTER;
    sep.exposureTime = (int)(100.0*exposure_i + 0.5);
    err=SBIGUnivDrvCommand(CC_START_EXPOSURE, &sep, NULL);   

    /* Did this work? */

    if ( err != CE_NO_ERROR ) 
    {
      fprintf(stderr,"Request to start imaging exposure ignored");
      return(0);
    } 
   
    /* Start guide exposure */
  
    sep.ccd = CCD_TRACKING;
    sep.abgState = ABG_LOW7;
    sep.openShutter =  SC_LEAVE_SHUTTER;
    sep.exposureTime = (int)(100.0*exposure_t + 0.5);
    err=SBIGUnivDrvCommand(CC_START_EXPOSURE, &sep, NULL);

    /* Did this work? */

    if ( err != CE_NO_ERROR ) 
    {
      fprintf(stderr,"Request to start tracking exposure ignored");
      return(0);
    } 

    /* Check and save tracking ccd bounds */
    
    if (subarea_t == FALSE)
    {
      x_t = 0;
      y_t = 0;
      width_t = ccd_track_width;
      height_t = ccd_track_height;
    }  
    else 
    {
      if (x_t < 0)
      {
        x_t = 0;
      }
      if (y_t < 0)
      {
        y_t = 0;
      }
      if (x_t > ccd_track_width  )
      {
        x_t = ccd_track_width;
      }
      if (y_t > ccd_track_height )
      {
        y_t = ccd_track_height;
      }
    }
    
    guide_x = x_t;
    guide_y = y_t;
    guide_w = width_t;
    guide_h = height_t;
    guide_n = guide_w*guide_h;       
    
    /* Center AO at the start of an exposure if OPTICAL */ 
    
    if  ( guide == OPTICAL )
    {
      SetCameraAOTipTilt(2048,2048); 
    }
    
    /* Return indicating that a guided exposure has been started               */

    ccd_phase = 1;
    *phase = ccd_phase;
    return(1);
  }    

  if ( ccd_phase != 1 )
  {
    /* We have already handled all allowed values except 1 */

    *phase = ccd_phase;
    return(0) ;     
  }  

  /* Are we done yet? */

  GetCameraStatus(&ccd_image_status, &ccd_track_status);
 
  /* Test for need to service tracking */
   
  if (ccd_track_status == COMPLETE)
  {    
    
    /* Transfer the track CCD data  */
    /* Any camera errors are noted but treated as non-fatal */ 
    
    eep.ccd = CCD_TRACKING;
    srp.ccd = CCD_TRACKING;
    rlp.ccd = CCD_TRACKING;
    err=SBIGUnivDrvCommand(CC_END_EXPOSURE, &eep, NULL);
    if ( err != CE_NO_ERROR ) 
    { 
      fprintf(stderr,"Error ending track ccd exposure during guided image");
      return(0);
    }  
    
    srp.readoutMode = 0;
    srp.top = guide_y;
    srp.left = guide_x;
    srp.width = guide_w;
    srp.height = guide_h;
    err = SBIGUnivDrvCommand(CC_START_READOUT, &srp, NULL);
    if ( err != CE_NO_ERROR ) 
    {
      fprintf(stderr,"Error starting track ccd readout during guided image");
      return(0);
    } 

    for (i = 0; i < srp.height; ++i) 
    {
      rlp.readoutMode = 0;
      rlp.pixelStart = guide_x;
      rlp.pixelLength = guide_w;
      err = SBIGUnivDrvCommand(CC_READOUT_LINE, &rlp, track_data + i*guide_w);
      if (err != CE_NO_ERROR) 
      {
        fprintf(stderr,"Error reading track ccd for guide star image");
        return(0);
      }
    }
                
    /* Use the latest tracking data to correct the pointing */
      
    GuideImage(track_data, 
      guide_w, guide_h, guide, guide_flag, guide_log_flag);
 
    /* Flush the buffers */
          
    fflush(stdout);
    sync();
    
    /* Start another track exposure if the image exposure is not complete */

    if ( ccd_image_status != COMPLETE )
    {    
            
      err=SBIGUnivDrvCommand(CC_START_EXPOSURE, &sep, NULL);
      if (err != CE_NO_ERROR) 
      {
        fprintf(stderr,"Error starting new track exposure");
        return(0);
      }
    }        

    /* Return with value indicating exposure still underway */
    
    ccd_phase = 1;
    *phase = ccd_phase; 
    return(1);
  } 
  
  /* Check on the image CCD and return if it is not complete */   
   
  if ( ccd_image_status != COMPLETE )
  {
    ccd_phase = 1;
    *phase = ccd_phase;    
    /* The exposure is still underway */
    return(1);
  }    

  /* Test for image subarea request and adjust parameters as needed */
 
  if (subarea_i == FALSE)
  {
     x_i = 0;
     y_i = 0;
     width_i = ccd_image_width;
     height_i = ccd_image_height;
  }  
  else 
  {
    if (x_i < 0)
    {
      x_i = 0;
    }
    if (y_i < 0)
    {
      y_i = 0;
    }
    if (x_i > ccd_image_width  )
    {
      x_i = ccd_image_width;
    }
    if (y_i > ccd_image_height )
    {
      y_i = ccd_image_height;
    }
  }          
  
  /* Transfer the CCD data to image_data */
  
  eep.ccd = CCD_IMAGING;
  srp.ccd = CCD_IMAGING;
  rlp.ccd = CCD_IMAGING;
  err=SBIGUnivDrvCommand(CC_END_EXPOSURE, &eep, NULL);
  if ( err != CE_NO_ERROR ) 
  {
    fprintf(stderr,"Error starting ending image exposure");
    return(0);
  } 
  
  /* Read it */

  srp.readoutMode = 0;
  srp.top = y_i;
  srp.left = x_i;
  srp.width = width_i;
  srp.height = height_i;
  err = SBIGUnivDrvCommand(CC_START_READOUT, &srp, NULL);
  if ( err != CE_NO_ERROR ) 
  {
    fprintf(stderr,"Error starting image exposure readout");
    return(0);
  } 
  
  for (i = 0; i < srp.height; ++i) 
  {
    rlp.readoutMode = 0;
    rlp.pixelStart = x_i;
    rlp.pixelLength = width_i;
    err = SBIGUnivDrvCommand(CC_READOUT_LINE,&rlp,image_data + i*width_i);
    if (err != CE_NO_ERROR) 
    {
      fprintf(stderr,"Error within image readout after guided exposure");
      return(0);
    }
  }

  /* Successful image readout ends guiding too */
  
  guide_phase = 0;
  
  /* Indicate that a new image is available */
  
  ccd_phase = 2;
  *phase = ccd_phase;  
  return(1);
}

/* Autoguiding */

/* Feedback to relays based on guide star error signal */
/* Global guide parameters deliver calibrated response */
/* Uses int relay 1 2 4 8 -> +X -X +Y -Y               */
/* Activates the relay servicing largest error         */
/* Make repeated calls to bring both errors down       */

void feedback_tracking(double x_err, double y_err)
{
  int relay;
  double relay_on_time;
  double fx, fy;
     
  fx = guide_x_to_track_x *x_err + guide_y_to_track_x*y_err;
  fy = guide_x_to_track_y *x_err + guide_y_to_track_y*y_err;
  
  printf("jfk: fx=%lf fy=%lf\n", fx, fy);
    
  if (fx*fx > fy*fy)
  {
    if ( fx > 0. )
    {
      relay = 1;
      relay_on_time = fx;  
      printf("jfk: relay = %d  time = %lf\n",relay, relay_on_time);
      ActivateCameraRelay(relay, relay_on_time);
    }
    else if ( fx < 0. )
    {
      relay = 2;
      relay_on_time = -fx; 
      printf("jfk: relay = %d  time = %lf\n",relay, relay_on_time);
      ActivateCameraRelay(relay, relay_on_time);
    }  
  }
  else
  {
    if ( fy > 0. )
    {
      relay = 4;
      relay_on_time = fy;  
      printf("jfk: relay = %d  time = %lf\n",relay, relay_on_time);
      ActivateCameraRelay(relay, relay_on_time);
    }
    else if ( fy < 0. )
    {
      relay = 8;
      relay_on_time = -fy;  
      printf("jfk: relay = %d  time = %lf\n",relay, relay_on_time);
      ActivateCameraRelay(relay, relay_on_time);
    }
  }  
  return;
}

/* Feedback ao system based on guide star offset       */
/* Global guide parameters deliver calibrated response */
/* AO center is (2048,2048)                            */ 
/* Error coordinate center is (0,0)                    */
/* Uses SetCameraAOTipTilt to check bounds             */

void feedback_ao(double x_err,double y_err)
{
  int ao_x, ao_y;
  double fx, fy;
   
  fx = guide_x_to_ao_x *x_err + guide_y_to_ao_x*y_err;
  fy = guide_x_to_ao_y *x_err + guide_y_to_ao_y*y_err;
  ao_x = 2048 + (int) fx;
  ao_y = 2048 + (int) fy;
  
  SetCameraAOTipTilt(ao_x,ao_y);
  
  return;
}
